from flowws import try_to_import

from .version import __version__

LocalDensity = try_to_import('.LocalDensity', 'LocalDensity', __name__)
RDF = try_to_import('.RDF', 'RDF', __name__)
SmoothBOD = try_to_import('.SmoothBOD', 'SmoothBOD', __name__)
Steinhardt = try_to_import('.Steinhardt', 'Steinhardt', __name__)



import collections

import flowws
from flowws import Argument as Arg
import freud
import numpy as np

@flowws.add_stage_arguments
class SmoothBOD(flowws.Stage):
    """Compute and display Bond Orientational Order Diagrams (BOODs)"""
    ARGS = [
        Arg('num_neighbors', '-n', int, default=4,
            help='Number of neighbors to compute'),
        Arg('use_distance', '-d', bool, default=False,
            help='Use distance, rather than num_neighbors, to find bonds'),
        Arg('r_max', type=float, default=2,
            help='Maximum radial distance if use_distance is given'),
        Arg('on_surface', type=bool, default=True,
            help='Restrict the BOOD to be on the surface of a sphere'),
        Arg('average', type=bool, default=False,
            help='If True, average the BOOD'),
        Arg('average_keys', type=[str],
            help='List of scope keys to generate distinct series when averaging'),
    ]

    def __init__(self, *args, **kwargs):
        self._data_cache = collections.defaultdict(list)
        self._run_cache_keys = set()
        super().__init__(*args, **kwargs)

    def run(self, scope, storage):
        """Compute the bonds in the system"""
        box = freud.box.Box.from_box(scope['box'])
        positions = scope['position']

        aq = freud.AABBQuery(box, positions)
        args = dict(num_neighbors=self.arguments['num_neighbors'],
                    exclude_ii=True, r_guess=self.arguments['r_max'])
        if self.arguments['use_distance']:
            args['mode'] = 'ball'
            args['r_max'] = self.arguments['r_max']

        nlist = aq.query(positions, args).toNeighborList()
        rijs = positions[nlist.point_indices] - positions[nlist.query_point_indices]
        bonds = box.wrap(rijs)

        key_names = self.arguments.get('average_keys', [])
        self._last_data_key = tuple(scope[name] for name in key_names)
        if self.arguments['average']:
            if scope.get('cache_key', object()) not in self._run_cache_keys:
                self._data_cache[self._last_data_key].append(bonds)
            if 'cache_key' in scope:
                self._run_cache_keys.add(scope['cache_key'])
        else:
            self._data_cache[self._last_data_key] = [bonds]

        scope['SmoothBOD.bonds'] = bonds
        scope.setdefault('visuals', []).append(self)
        scope.setdefault('visual_link_rotation', []).append(self)

    def draw_plato(self):
        import plato, plato.draw as draw
        bonds = np.concatenate(self._data_cache[self._last_data_key], axis=0)
        prim = draw.SpherePoints(points=bonds, on_surface=self.arguments['on_surface'])
        scene = draw.Scene(prim, size=(3, 3), pixel_scale=100,
                           features=dict(additive_rendering=dict(invert=True)))
        return scene



import flowws
from flowws import Argument as Arg
import freud

@flowws.add_stage_arguments
class LocalDensity(flowws.Stage):
    """Compute the local density of particles in the system"""
    ARGS = [
        Arg('histogram_bins', '-b', int, default=64,
            help='Number of bins to use in the histogram plot'),
        Arg('r_max', '-r', float, required=True,
            help='Maximum radial distance'),
        Arg('diameter', '-d', float, default=0.,
            help='Smoothing diameter to use in the density calculation'),
    ]

    def run(self, scope, storage):
        """Compute and provide the local density"""
        compute = freud.density.LocalDensity(
            self.arguments['r_max'],
            self.arguments['diameter'])

        box = freud.box.Box.from_box(scope['box'], scope.get('dimensions', 3))
        compute.compute((box, scope['position']), scope['position'])
        self.density = scope['local_density'] = compute.density

        scope.setdefault('color_scalars', []).append('local_density')
        scope.setdefault('visuals', []).append(self)

    def draw_matplotlib(self, figure):
        ax = figure.add_subplot(111)
        ax.hist(self.density, bins=self.arguments['histogram_bins'],
                density=True)
        ax.set_xlabel('Density')
        ax.set_ylabel('Probability')



import flowws
from flowws import Argument as Arg
import freud

@flowws.add_stage_arguments
class Steinhardt(flowws.Stage):
    """Compute the Steinhardt order parameter of particles in the system"""
    ARGS = [
        Arg('histogram_bins', '-b', int, default=64,
            help='Number of bins to use in the histogram plot'),
        Arg('l', '-l', int, default=6,
            help='Spherical harmonic degree of the order parameter'),
        Arg('r_max', '-r', float,
            help='Maximum radial distance to consider for neighbors (if given)'),
        Arg('num_neighbors', '-n', int,
            help='Number of neighbors to use; overrules r_max if given'),
        Arg('r_guess', None, float, 2,
            help='Characteristic distance for finding num_neighbors neighboring particles'),
    ]

    def run(self, scope, storage):
        """Compute and provide the Steinhardt order parameter"""
        compute = freud.order.Steinhardt(self.arguments['l'])

        box = freud.box.Box.from_box(scope['box'], scope.get('dimensions', 3))
        query_options = dict(exclude_ii=True)
        if 'num_neighbors' in self.arguments:
            query_options['num_neighbors'] = self.arguments['num_neighbors']
            query_options['r_guess'] = self.arguments['r_guess']
        else:
            query_options['r_max'] = self.arguments.get('r_max', None)
        compute.compute((box, scope['position']), query_options)

        self.steinhardt = compute.ql
        name = 'steinhardt_q{}'.format(self.arguments['l'])
        scope[name] = self.steinhardt
        scope.setdefault('color_scalars', []).append(name)
        scope.setdefault('visuals', []).append(self)

    def draw_matplotlib(self, figure):
        ax = figure.add_subplot(111)
        ax.hist(self.steinhardt, bins=self.arguments['histogram_bins'],
                density=True)
        ax.set_xlabel('$Q_{{{}}}$'.format(self.arguments['l']))
        ax.set_ylabel('Probability')



#!/usr/bin/env python

import os
from setuptools import setup

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

version_fname = os.path.join(THIS_DIR, 'flowws_freud', 'version.py')
with open(version_fname) as version_file:
    exec(version_file.read())

readme_fname = os.path.join(THIS_DIR, 'README.md')
with open(readme_fname) as readme_file:
    long_description = readme_file.read()

module_names = [
    'LocalDensity',
    'RDF',
    'SmoothBOD',
    'Steinhardt',
]

flowws_modules = []
for name in module_names:
    flowws_modules.append('{0} = flowws_freud.{0}:{0}'.format(name))
    flowws_modules.append(
        'flowws_freud.{0} = flowws_freud.{0}:{0}'.format(name))

setup(name='flowws-freud',
      author='Matthew Spellings',
      author_email='matthew.p.spellings@gmail.com',
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3',
      ],
      description='Stage-based scientific workflows using freud',
      entry_points={
          'flowws_modules': flowws_modules,
      },
      extras_require={},
      install_requires=[
          'flowws',
          'freud-analysis',
      ],
      license='MIT',
      long_description=long_description,
      long_description_content_type='text/markdown',
      packages=[
          'flowws_freud',
      ],
      python_requires='>=3',
      version=__version__
      )



import flowws
from flowws import Argument as Arg
import freud

@flowws.add_stage_arguments
class RDF(flowws.Stage):
    """Compute and plot the radial distribution function (RDF)"""
    ARGS = [
        Arg('bins', '-b', int, default=64,
            help='Number of bins to use'),
        Arg('r_min', type=float, default=0,
            help='Minimum radial distance'),
        Arg('r_max', type=float, required=True,
            help='Maximum radial distance'),
        Arg('bond_max', type=float, default=0,
            help='Render bonds that have length up to the given distance'),
        Arg('bond_width', type=float, default=.1,
            help='Width of drawn bonds, if enabled'),
    ]

    def run(self, scope, storage):
        """Compute and provide the RDF"""
        compute = freud.density.RDF(
            self.arguments['bins'], self.arguments['r_max'],
            self.arguments.get('r_min', 0))

        box = freud.box.Box.from_box(scope['box'])
        system = freud.locality.AABBQuery(box, scope['position'])
        compute.compute(system)
        self.r, self.rdf = compute.bin_centers, compute.rdf

        if self.arguments.get('bond_max', None):
            import plato.draw

            query_args = dict(
                mode='ball', r_max=self.arguments['bond_max'], exclude_ii=True)
            query = system.query(scope['position'], query_args)
            nlist = query.toNeighborList()

            start_points = scope['position'][nlist.query_point_indices]
            bonds = (scope['position'][nlist.point_indices] - start_points)
            bonds = box.wrap(bonds)
            end_points = start_points + bonds

            prim = plato.draw.Lines(
                start_points=start_points, end_points=end_points,
                widths=self.arguments['bond_width'], colors=(.1, .1, .1, 1),
                cap_mode=1)
            scope.setdefault('plato_primitives', []).append(prim)

        scope.setdefault('visuals', []).append(self)

    def draw_matplotlib(self, figure):
        ax = figure.add_subplot(111)
        ax.plot(self.r, self.rdf)
        ax.set_xlabel('r')
        ax.set_ylabel('RDF')
